# -*- coding: utf-8 -*-
"""
w2lapp.passwd.py: change password associated with entry

web2ldap - a web-based LDAP Client,
see http://www.web2ldap.de for details

(c) by Michael Stroeder <michael@stroeder.com>

This module is distributed under the terms of the
GPL (GNU GENERAL PUBLIC LICENSE) Version 2
(see http://www.gnu.org/copyleft/gpl.html)

$Id: passwd.py,v 1.67 2010/01/10 17:10:30 michael Exp $
"""

import time,binascii,hashlib,ldap,ldaputil.base,ldaputil.passwd, \
       w2lapp.cnf,w2lapp.core,w2lapp.gui,w2lapp.login

available_hashtypes = ldaputil.passwd.AVAIL_USERPASSWORD_SCHEMES.items()


PASSWD_ACTIONS = [
  (u'passwdextop',u'Server-side',u'Password modify extended operation'),
  (u'setuserpassword',u'Modify password attribute',u'Set the password attribute with modify operation'),
#  ('deluserpassword',u'Del. userPassword',u'Delete the attribute userPassword'),
#  ('setunicodepwd',u'Set AD password (unicodePwd)',u'Set AD password (unicodePwd)'),
]
PASSWD_ACTIONS_DICT = dict([
  (action,(short_desc,long_desc))
  for action,short_desc,long_desc in PASSWD_ACTIONS
])

def NtPasswordHash(password):
  """
  Returns MD4-based NT password hash for a password.
  """
  md4_digest = hashlib.new('md4')
  md4_digest.update(password[:128].encode('utf-16-le'))
  return binascii.hexlify(md4_digest.digest()).upper()


def GetAllAllowedAttributes(schema,oc_list):
  r,a = schema.attribute_types(oc_list,raise_keyerror=0)
  result = {}
  result.update(r)
  result.update(a)
  return result # GetAllAllowedAttributes()


def PasswdContextMenu(sid,form,ls,dn):
  return [
    form.applAnchor(
      'passwd',
      short_desc,sid,
      [
        ('dn',dn),
        ('passwd_action',pa),
        ('passwd_who',dn),
      ],
      title=long_desc
    )
    for pa,short_desc,long_desc in PASSWD_ACTIONS
  ]


def UnicodePwdAvailable(ls,dn,sub_schema):
  return sub_schema.sed[ldap.schema.AttributeType].has_key('1.2.840.113556.1.4.90')


def OwnPwdChange(ls,dn):
  return (ls.who is None) or (ls.who==dn)


def PasswdForm(
  sid,outf,command,form,ls,dn,sub_schema,
  passwd_action,passwd_who,user_objectclasses,
  Heading,Msg
):

  # Determine whether password attribute is unicodePwd on MS AD
  unicode_pwd_avail = UnicodePwdAvailable(ls,dn,sub_schema)
  
  # Determine whether user changes own password
  own_pwd_change = OwnPwdChange(ls,passwd_who)

  all_attrs_dict = GetAllAllowedAttributes(sub_schema,user_objectclasses)

  if not unicode_pwd_avail:

  #  config_hashtypes = ls.rootDSE.get('supportedAuthPasswordSchemes',None) or w2lapp.cnf.GetParam(ls,'passwd_hashtypes',[])
    config_hashtypes = w2lapp.cnf.GetParam(ls,'passwd_hashtypes',[])
    if config_hashtypes:
      # The set of hash types are restricted by local configuration
      default_hashtypes = [
        hash_type
        for hash_type in available_hashtypes
        if hash_type[0] in config_hashtypes
      ]
      form.field['passwd_scheme'].options = default_hashtypes

    # Check for ntPassword and sambaNTPassword attribute
    if all_attrs_dict.has_key('1.3.6.1.4.1.7165.2.1.2') or \
       all_attrs_dict.has_key('1.3.6.1.4.1.7165.2.1.25'):
      ntpassword_sync_field = form.field['passwd_ntpasswordsync'].inputHTML()
      ntpassword_sync_tablerow = """<tr>
        <td>Sync with NtPassword attribute (for Samba)</td>
        <td>%s</td>
      </tr>
      """ % (ntpassword_sync_field)
    else:
      ntpassword_sync_tablerow = ''

  w2lapp.gui.TopSection(
    sid,outf,form,ls,dn,'Change password',
    w2lapp.gui.MainMenu(sid,form,ls,dn),
    context_menu_list=PasswdContextMenu(sid,form,ls,dn)
  )

  if own_pwd_change:
    old_passwd_input_html = """
      <tr>
        <td>Enter old password</td>
        <td>
          <input
            type="password"
            name="passwd_oldpasswd"
            maxlength="100"
            size="30"
            title="Your old password"
            value="">
        </td>
      </tr>
    """
  else:
    old_passwd_input_html = ''
    
  outf.write("""<div id="Input" class="Main">
    <h1>%s</h1>
    <p>%s</p>
    %s
    <input type="hidden" name="dn" value="%s">
    <input type="hidden" name="passwd_action" value="%s">
    <input type="hidden" name="passwd_who" value="%s">
    <table summary="Password input form">
    <tr>
      <td>%s</td>
      <td>%s</td>
    </tr>
    %s
    <tr>
      <td>Enter new password</td>
      <td>
        <input
          type="password"
          name="passwd_newpasswd"
          maxlength="100"
          size="30"
          title="The new password to be set"
          value="">
      </td>
    </tr>
    <tr>
      <td>Repeat new password</td>
      <td>
        <input
          type="password"
          name="passwd_newpasswd"
          maxlength="100"
          size="30"
          title="Again the new password for verification"
          value="">
      </td>
    </tr>
    """ % (
      Heading,
      form.utf2display(Msg or PASSWD_ACTIONS_DICT[passwd_action][1]),
      form.beginFormHTML('passwd',sid,'POST'),
      form.utf2display(dn,sp_entity='  '),
      form.utf2display(passwd_action,sp_entity='  '),
      form.utf2display(passwd_who,sp_entity='  '),
      {False:'Change password for',True:'Change own password'}[own_pwd_change],
      w2lapp.gui.DisplayDN(sid,form,ls,passwd_who),
      old_passwd_input_html,
  ))

  if passwd_action=='setuserpassword' and not unicode_pwd_avail:
    outf.write("""
      <tr>
        <td>Password encryption method</td>
        <td>%s</td>
      </tr>
      %s
      <tr>
        <td>Sync password setting times</td>
        <td>%s</td>
      </tr>
      """ % (
        form.field['passwd_scheme'].inputHTML(),
        ntpassword_sync_tablerow,
        form.field['passwd_settimesync'].inputHTML(checked=ls.who!=passwd_who),
    ))

  outf.write("""
      </table>
      <input type="submit" value="Change Password">
      <input type="reset" value="Reset">
      </form>
      </div>
  """)

  w2lapp.gui.PrintFooter(outf,form)
  return # PasswdForm()


def w2l_Passwd(sid,outf,command,form,ls,dn):

  sub_schema = ls.retrieveSubSchema(dn,w2lapp.cnf.GetParam(ls,'_schema',None))
  
  # Determine the default value for passwd_action based on
  # server's visible configuration
  if '1.3.6.1.4.1.4203.1.11.1' in ls.supportedExtension:
    # Password Modify extended operation seems to be announced in rootDSE
    passwd_action_default = u'passwdextop'
  else:
    passwd_action_default = u'setuserpassword'

  passwd_action = form.getInputValue('passwd_action',[passwd_action_default])[0]
  passwd_who = form.getInputValue('passwd_who',[dn])[0]

  user_objectclasses = [ oc.lower() for oc in ls.getObjectClasses(passwd_who)[0] ]

  if 'passwd_newpasswd' in form.inputFieldNames:

    ####################################################################
    # New password provided => (re)set it in entry
    ####################################################################

    if len(form.field['passwd_newpasswd'].value)!=2:
      raise w2lapp.core.ErrorExit(u'Repeat password!')

    if form.field['passwd_newpasswd'].value[0]!=form.field['passwd_newpasswd'].value[1]:
      PasswdForm(
        sid,outf,command,form,ls,dn,sub_schema,passwd_action,passwd_who,user_objectclasses,
        Heading='Password Error',Msg=u'New passwords do not match!'
      )
      return

    old_password = form.getInputValue('passwd_oldpasswd',[None])[0]

    passwd_input = form.field['passwd_newpasswd'].value[0]

    if not passwd_input:
      PasswdForm(
        sid,outf,command,form,ls,dn,sub_schema,passwd_action,passwd_who,user_objectclasses,
        Heading='Password Error',Msg=u'Empty passwords not allowed!'
      )
      return

    password_attr_types_msg = ''

    if passwd_action=='passwdextop':

      #-------------------------------------------------------
      # Modify password via Password Modify Extended Operation
      #-------------------------------------------------------

      try:
        ls.l.passwd_s(
          passwd_who.encode(ls.charset),
          (old_password or u'').encode(ls.charset) or None,
          passwd_input.encode(ls.charset)
        )
      except (ldap.CONSTRAINT_VIOLATION,ldap.UNWILLING_TO_PERFORM),e:
        PasswdForm(
          sid,outf,command,form,ls,dn,sub_schema,
          passwd_action,passwd_who,user_objectclasses,
          Heading='Password Error',Msg=w2lapp.gui.LDAPError2ErrMsg(e)
        )
        return
      else:
        password_attr_types_msg = 'Password changed at the server.'
    
    elif passwd_action in ('setuserpassword','setunicodepwd'):

      #-------------------------------------------------------
      # Modify password via Modify Request
      #-------------------------------------------------------

      modlist = []

      all_attrs_dict = GetAllAllowedAttributes(sub_schema,user_objectclasses)

      if not all_attrs_dict.has_key('2.5.4.35'):
        # Add a simple AUXILIARY object class for the userPassword attribute if 
        # current object classes of entry do not allow userPassword
        for userpassword_class in ('simpleSecurityObject','simpleAuthObject'):
          try:
            if not userpassword_class.lower() in user_objectclasses and \
               sub_schema.get_inheritedattr(ldap.schema.ObjectClass,userpassword_class,'kind')==2:
              modlist.append((ldap.MOD_ADD,'objectClass',userpassword_class))
              break
          except KeyError:
            pass

      passwd_scheme = form.getInputValue('passwd_scheme',[''])[0]

      # Set "standard" password of LDAP entry
      if all_attrs_dict.has_key('1.2.840.113556.1.4.90'):
        # Active Directory's password attribute unicodePwd
        passwd_attr_type = 'unicodePwd'
        unicodePwd = ldaputil.passwd.UnicodePwd(
          l=ls.l,
          dn=passwd_who.encode(ls.charset)
        )
        new_passwd_value = unicodePwd.encodePassword(passwd_input)
        passwd_scheme = ''
        if old_password:
          old_passwd_value = unicodePwd.encodePassword(old_password)
      else:
        # Assume standard password attribute userPassword
        passwd_attr_type = 'userPassword'
        userPassword = ldaputil.passwd.UserPassword(
          l=ls.l,
          dn=passwd_who.encode(ls.charset),
          charset=ls.charset,
          multiple=0
        )
        # Work around the problem that LDAP applications use different
        # charset encodings for the userPassword attribute
        new_passwd_value = userPassword.encodePassword(passwd_input,passwd_scheme)
        if old_password:
          old_passwd_value = userPassword.encodePassword(old_password,'')

      if OwnPwdChange(ls,passwd_who) and old_password:
        modlist.extend((
          (ldap.MOD_DELETE,passwd_attr_type,[old_passwd_value]),
          (ldap.MOD_ADD,passwd_attr_type,[new_passwd_value]),
        ))
      else:
        modlist.append(
          (ldap.MOD_REPLACE,passwd_attr_type,[new_passwd_value]),
        )

      passwd_settimesync = form.getInputValue('passwd_settimesync',['no'])[0]=='yes'

      pwd_change_timestamp = time.time()

      if passwd_settimesync and all_attrs_dict.has_key('1.3.6.1.1.1.1.5'):
        modlist.append((ldap.MOD_REPLACE,'shadowLastChange',str(int(pwd_change_timestamp/86400))))

      passwd_ntpasswordsync = form.getInputValue('passwd_ntpasswordsync',['no'])[0]=='yes'

      # Samba password synchronization if requested
      if passwd_ntpasswordsync:

        if all_attrs_dict.has_key('1.3.6.1.4.1.7165.2.1.2'):
          modlist.append((ldap.MOD_REPLACE,'ntPassword',NtPasswordHash(passwd_input)))

        if all_attrs_dict.has_key('1.3.6.1.4.1.7165.2.1.25'):
          modlist.append((ldap.MOD_REPLACE,'sambaNTPassword',NtPasswordHash(passwd_input)))

      if passwd_settimesync and passwd_ntpasswordsync:

        if all_attrs_dict.has_key('1.3.6.1.4.1.7165.2.1.3'):
          modlist.append((ldap.MOD_REPLACE,'pwdLastSet',str(int(pwd_change_timestamp))))

        if all_attrs_dict.has_key('1.3.6.1.4.1.7165.2.1.27'):
          modlist.append((ldap.MOD_REPLACE,'sambaPwdLastSet',str(int(pwd_change_timestamp))))

      password_attr_types_msg = 'Password-related attributes set:<br>\n<ul>%s</ul>' % ('\n'.join(
        [ 
          '<li>%s</li>' % (attr_type)
          for mod, attr_type, attr_values in modlist
        ]
      ))

      # Modify password
      try:
        ls.modifyEntry(passwd_who,modlist)
      except (ldap.CONSTRAINT_VIOLATION,ldap.UNWILLING_TO_PERFORM),e:
        PasswdForm(
          sid,outf,command,form,ls,dn,sub_schema,
          passwd_action,passwd_who,user_objectclasses,
          Heading='Password Error',Msg=w2lapp.gui.LDAPError2ErrMsg(e)
        )
        return
      except ldap.NO_SUCH_ATTRIBUTE,e:
        PasswdForm(
          sid,outf,command,form,ls,dn,sub_schema,
          passwd_action,passwd_who,user_objectclasses,
          Heading='Password Error',Msg=w2lapp.gui.LDAPError2ErrMsg(e)+u" (Hint: Try without old password.)"
        )
        return


    # Check if relogin is necessary
    if OwnPwdChange(ls,passwd_who) and old_password:
      ls.bind('','',None,'','')
      w2lapp.login.w2l_Login(
        sid,outf,'searchform',form,ls,dn,None,ls.getSearchRoot(passwd_who),
        login_msg=u'Password changed. A new login is required!',
        who=passwd_who,relogin=0,nomenu=1
      )
    else:
      w2lapp.gui.SimpleMessage(
        sid,outf,form,ls,dn,
        message='Changed password of entry %s.<br>\n%s' % (
          w2lapp.gui.DisplayDN(sid,form,ls,passwd_who),
          password_attr_types_msg,
        ),
        main_menu_list=w2lapp.gui.MainMenu(sid,form,ls,dn),
        context_menu_list=w2lapp.gui.ContextMenuSingleEntry(sid,form,ls,dn)
      )

  else:

    ####################################################################
    # New password not yet provided => ask for it
    ####################################################################

    PasswdForm(
      sid,outf,command,form,ls,dn,sub_schema,
      passwd_action,passwd_who,
      user_objectclasses,'Enter new password',''
    )
